@@ -14,7 +14,6 @@ import sys
 import os
 
 from gevent.hub import _get_hub_noargs as get_hub
-from gevent._compat import PY2
 from gevent._compat import integer_types
 from gevent._compat import reraise
 from gevent._compat import fspath
@@ -393,18 +392,12 @@ class OpenDescriptor(object): # pylint:disable=too-many-instance-attributes
 
         if not self.binary:
             # Either native or text at this point.
-            if PY2 and self.native:
-                # Neither text mode nor binary mode specified.
-                if self.universal:
-                    # universal was requested, e.g., 'rU'
-                    result = UniversalNewlineBytesWrapper(result, line_buffering)
-            else:
-                # Python 2 and text mode, or Python 3 and either text or native (both are the same)
-                if not isinstance(raw, io.TextIOBase):
-                    # Avoid double-wrapping a TextIOBase in another TextIOWrapper.
-                    # That tends not to work. See https://github.com/gevent/gevent/issues/1542
-                    result = io.TextIOWrapper(result, self.encoding, self.errors, self.newline,
-                                              line_buffering)
+            # Python 2 and text mode, or Python 3 and either text or native (both are the same)
+            if not isinstance(raw, io.TextIOBase):
+                # Avoid double-wrapping a TextIOBase in another TextIOWrapper.
+                # That tends not to work. See https://github.com/gevent/gevent/issues/1542
+                result = io.TextIOWrapper(result, self.encoding, self.errors, self.newline,
+                                          line_buffering)
 
         if result is not raw or self._raw_object_is_new(raw):
             # Set the mode, if possible, but only if we created a new
@@ -428,6 +421,30 @@ class OpenDescriptor(object): # pylint:disable=too-many-instance-attributes
         return result
 
 
+class _ClosedIO(object):
+    # Used for FileObjectBase._io when FOB.close()
+    # is called. Lets us drop references to ``_io``
+    # for GC/resource cleanup reasons, but keeps some useful
+    # information around.
+    __slots__ = ('name',)
+
+    def __init__(self, io_obj):
+        try:
+            self.name = io_obj.name
+        except AttributeError:
+            pass
+
+    def __getattr__(self, name):
+        if name == 'name':
+            # We didn't set it in __init__ because there wasn't one
+            raise AttributeError
+        raise FileObjectClosed
+
+    def __bool__(self):
+        return False
+    __nonzero__ = __bool__
+
+
 class FileObjectBase(object):
     """
     Internal base class to ensure a level of consistency
@@ -452,6 +469,7 @@ class FileObjectBase(object):
         'readline',
         'readlines',
         'read1',
+        'readinto',
 
         # Write.
         # Note that we do not extend WriteallMixin,
@@ -472,7 +490,6 @@ class FileObjectBase(object):
         # We don't actually use this property ourself, but we save it (and
         # pass it along) for compatibility.
         self._close = descriptor.closefd
-
         self._do_delegate_methods()
 
 
@@ -503,14 +520,14 @@ class FileObjectBase(object):
     @property
     def closed(self):
         """True if the file is closed"""
-        return self._io is None
+        return isinstance(self._io, _ClosedIO)
 
     def close(self):
-        if self._io is None:
+        if isinstance(self._io, _ClosedIO):
             return
 
         fobj = self._io
-        self._io = None
+        self._io = _ClosedIO(self._io)
         try:
             self._do_close(fobj, self._close)
         finally:
@@ -525,8 +542,6 @@ class FileObjectBase(object):
         raise NotImplementedError()
 
     def __getattr__(self, name):
-        if self._io is None:
-            raise FileObjectClosed()
         return getattr(self._io, name)
 
     def __repr__(self):
@@ -656,8 +671,6 @@ class FileObjectThread(FileObjectBase):
 
     def _do_delegate_methods(self):
         FileObjectBase._do_delegate_methods(self)
-        # if not hasattr(self, 'read1') and 'r' in getattr(self._io, 'mode', ''):
-        #     self.read1 = self.read
         self.__io_holder[0] = self._io
 
     def _extra_repr(self):
@@ -676,7 +689,7 @@ class FileObjectThread(FileObjectBase):
                 # This is different than FileObjectPosix, etc,
                 # because we want to save the expensive trip through
                 # the threadpool.
-                raise FileObjectClosed()
+                raise FileObjectClosed
             with lock:
                 return threadpool.apply(method, args, kwargs)
 
