@@ -5,6 +5,7 @@ import re
 import sys
 import os
 import glob
+import operator
 import traceback
 import importlib
 
@@ -33,7 +34,7 @@ from . import travis
 # See https://bitbucket.org/pypy/pypy/issues/2769/systemerror-unexpected-internal-exception
 try:
     __import__('_testcapi')
-except (ImportError, OSError, IOError):
+except (ImportError, OSError):
     # This can raise a wide variety of errors
     pass
 
@@ -80,8 +81,13 @@ class ResultCollector(object):
         self.passed = {}
         self.total_cases = 0
         self.total_skipped = 0
+        # Every RunResult reported: failed, passed, rerun
+        self._all_results = []
+        self.reran = {}
 
     def __iadd__(self, result):
+        self._all_results.append(result)
+
         if not result:
             self.failed[result.name] = result #[cmd, kwargs]
         else:
@@ -90,6 +96,30 @@ class ResultCollector(object):
         self.total_skipped += result.skipped_count
         return self
 
+    def __ilshift__(self, result):
+        """
+        collector <<= result
+
+        Stores the result, but does not count it towards
+        the number of cases run, skipped, passed or failed.
+        """
+        self._all_results.append(result)
+        self.reran[result.name] = result
+        return self
+
+    @property
+    def longest_running_tests(self):
+        """
+        A new list of RunResult objects, sorted from longest running
+        to shortest running.
+        """
+        return sorted(self._all_results,
+                      key=operator.attrgetter('run_duration'),
+                      reverse=True)
+
+
+class FailFast(Exception):
+    pass
 
 class Runner(object):
 
@@ -102,7 +132,8 @@ class Runner(object):
                  failfast=False,
                  quiet=False,
                  configured_run_alone_tests=(),
-                 worker_count=DEFAULT_NWORKERS):
+                 worker_count=DEFAULT_NWORKERS,
+                 second_chance=False):
         """
         :keyword quiet: Set to True or False to explicitly choose. Set to
             `None` to use the default, which may come from the environment variable
@@ -110,10 +141,13 @@ class Runner(object):
         """
         self._tests = tests
         self._configured_failing_tests = configured_failing_tests
-        self._failfast = failfast
         self._quiet = quiet
         self._configured_run_alone_tests = configured_run_alone_tests
 
+        assert not (failfast and second_chance)
+        self._failfast = failfast
+        self._second_chance = second_chance
+
         self.results = ResultCollector()
         self.results.total = len(self._tests)
         self._running_jobs = []
@@ -124,8 +158,16 @@ class Runner(object):
         if self._quiet is not None:
             kwargs['quiet'] = self._quiet
         result = util.run(cmd, **kwargs)
+        if not result and self._second_chance:
+            self.results <<= result
+            util.log("> %s", result.name, color='warning')
+            result = util.run(cmd, **kwargs)
         if not result and self._failfast:
-            sys.exit(1)
+            # Under Python 3.9 (maybe older versions?), raising the
+            # SystemExit here (a background thread belonging to the
+            # pool) doesn't seem to work well. It gets stuck waiting
+            # for a lock? The job never shows up as finished.
+            raise FailFast(cmd)
         self.results += result
 
     def _reap(self):
@@ -141,7 +183,10 @@ class Runner(object):
         return len(self._running_jobs)
 
     def _reap_all(self):
-        while self._reap() > 0:
+        util.log("Reaping %d jobs", len(self._running_jobs), color="debug")
+        while self._running_jobs:
+            if not self._reap():
+                break
             util.sleep(self.TIME_WAIT_REAP)
 
     def _spawn(self, pool, cmd, options):
@@ -211,12 +256,10 @@ class Runner(object):
     def _report(self, elapsed_time, exit=False):
         results = self.results
         report(
-            results.total, results.failed, results.passed,
+            results,
             exit=exit,
             took=elapsed_time,
             configured_failing_tests=self._configured_failing_tests,
-            total_cases=results.total_cases,
-            total_skipped=results.total_skipped
         )
 
 
@@ -368,9 +411,27 @@ class Discovery(object):
 
         def __begin_command(self):
             cmd = [sys.executable, '-u']
-            if PYPY and PY2:
-                # Doesn't seem to be an env var for this
-                cmd.extend(('-X', 'track-resources'))
+            # XXX: -X track-resources is broken. This happened when I updated to
+            # PyPy 7.3.2. It started failing to even start inside the virtual environment
+            # with
+            #
+            # debug: OperationError:
+            # debug:  operror-type: ImportError
+            # debug:  operror-value: No module named traceback
+            #
+            # I don't know if this is PyPy's problem or a problem in virtualenv:
+            #
+            # virtualenv==20.0.35
+            # virtualenv-clone==0.5.4
+            # virtualenvwrapper==4.8.4
+            #
+            # Deferring investigation until I need this...
+
+            # if PYPY and PY2:
+            #     # Doesn't seem to be an env var for this.
+            #     # XXX: track-resources is broken in virtual environments
+            #     # on 7.3.2.
+            #     cmd.extend(('-X', 'track-resources'))
             return cmd
 
 
@@ -454,7 +515,15 @@ class Discovery(object):
                 module_name = os.path.splitext(filename)[0]
                 qualified_name = self.package + '.' + module_name if self.package else module_name
 
-            with open(os.path.abspath(filename), 'rb') as f:
+            # Also allow just 'foo' as a shortcut for 'gevent.tests.foo'
+            abs_filename = os.path.abspath(filename)
+            if (
+                    not os.path.exists(abs_filename)
+                    and not filename.endswith('.py')
+                    and os.path.exists(abs_filename + '.py') ):
+                abs_filename += '.py'
+
+            with open(abs_filename, 'rb') as f:
                 # Some of the test files (e.g., test__socket_dns) are
                 # UTF8 encoded. Depending on the environment, Python 3 may
                 # try to decode those as ASCII, which fails with UnicodeDecodeError.
@@ -520,6 +589,7 @@ class Discovery(object):
 def load_list_from_file(filename, package):
     result = []
     if filename:
+        # pylint:disable=unspecified-encoding
         with open(_package_relative_filename(filename, package)) as f:
             for x in f:
                 x = x.split('#', 1)[0].strip()
@@ -555,18 +625,38 @@ def format_seconds(seconds):
     return seconds
 
 
-def report(total, failed, passed, exit=True, took=None,
-           configured_failing_tests=(),
-           total_cases=0, total_skipped=0):
+def _show_longest_running(result_collector, how_many=5):
+    longest_running_tests = result_collector.longest_running_tests
+    if not longest_running_tests:
+        return
+    # The only tricky part is handling repeats. we want to show them,
+    # but not count them as a distinct entry.
+
+    util.log('\nLongest-running tests:')
+    length_of_longest_formatted_decimal = len('%.1f' % longest_running_tests[0].run_duration)
+
+    frmt = '%' + str(length_of_longest_formatted_decimal) + '.1f seconds: %s'
+    seen_names = set()
+    for result in longest_running_tests:
+        util.log(frmt, result.run_duration, result.name)
+        seen_names.add(result.name)
+        if len(seen_names) >= how_many:
+            break
+
+
+
+def report(result_collector, # type: ResultCollector
+           exit=True, took=None,
+           configured_failing_tests=()):
     # pylint:disable=redefined-builtin,too-many-branches,too-many-locals
-    runtimelog = util.runtimelog # XXX: Global state!
-    if runtimelog:
-        util.log('\nLongest-running tests:')
-        runtimelog.sort()
-        length = len('%.1f' % -runtimelog[0][0])
-        frmt = '%' + str(length) + '.1f seconds: %s'
-        for delta, name in runtimelog[:5]:
-            util.log(frmt, -delta, name)
+    total = result_collector.total
+    failed = result_collector.failed
+    passed = result_collector.passed
+    total_cases = result_collector.total_cases
+    total_skipped = result_collector.total_skipped
+
+    _show_longest_running(result_collector)
+
     if took:
         took = ' in %s' % format_seconds(took)
     else:
@@ -584,6 +674,10 @@ def report(total, failed, passed, exit=True, took=None,
         util.log('\n%s/%s unexpected passes', len(passed_unexpected), total, color='error')
         print_list(passed_unexpected)
 
+    if result_collector.reran:
+        util.log('\n%s/%s tests rerun', len(result_collector.reran), total, color='warning')
+        print_list(result_collector.reran)
+
     if failed:
         util.log('\n%s/%s tests failed%s', len(failed), total, took, color='warning')
 
@@ -623,14 +717,31 @@ def print_list(lst):
         util.log(' - %s', name)
 
 def _setup_environ(debug=False):
-    if ('PYTHONWARNINGS' not in os.environ
+    def not_set(key):
+        return not bool(os.environ.get(key))
+
+    if (not_set('PYTHONWARNINGS')
             and (not sys.warnoptions
                  # Python 3.7 goes from [] to ['default'] for nothing
                  or sys.warnoptions == ['default'])):
         # action:message:category:module:line
-        os.environ['PYTHONWARNINGS'] = ','.join([
-            # Enable default warnings such as ResourceWarning.
+
+        # - when a warning matches
+        #   more than one option, the action for the last matching
+        #   option is performed.
+        # - action is one of : ignore, default, all, module, once, error
+
+        # Enable default warnings such as ResourceWarning.
+        # ResourceWarning doesn't exist on Py2, so don't put it
+        # in there to avoid a warnnig.
+        defaults = [
             'default',
+            'default::DeprecationWarning',
+        ]
+        if not PY2:
+            defaults.append('default::ResourceWarning')
+
+        os.environ['PYTHONWARNINGS'] = ','.join(defaults + [
             # On Python 3[.6], the system site.py module has
             # "open(fullname, 'rU')" which produces the warning that
             # 'U' is deprecated, so ignore warnings from site.py
@@ -655,22 +766,22 @@ def _setup_environ(debug=False):
             'ignore:::dns.zone:',
         ])
 
-    if 'PYTHONFAULTHANDLER' not in os.environ:
+    if not_set('PYTHONFAULTHANDLER'):
         os.environ['PYTHONFAULTHANDLER'] = 'true'
 
-    if 'GEVENT_DEBUG' not in os.environ and debug:
+    if not_set('GEVENT_DEBUG') and debug:
         os.environ['GEVENT_DEBUG'] = 'debug'
 
-    if 'PYTHONTRACEMALLOC' not in os.environ and debug:
+    if not_set('PYTHONTRACEMALLOC') and debug:
         # This slows the tests down quite a bit. Reserve
         # for debugging.
         os.environ['PYTHONTRACEMALLOC'] = '10'
 
-    if 'PYTHONDEVMODE' not in os.environ:
+    if not_set('PYTHONDEVMODE'):
         # Python 3.7 and above.
         os.environ['PYTHONDEVMODE'] = '1'
 
-    if 'PYTHONMALLOC' not in os.environ and debug:
+    if not_set('PYTHONMALLOC') and debug:
         # Python 3.6 and above.
         # This slows the tests down some, but
         # can detect memory corruption. Unfortunately
@@ -682,21 +793,54 @@ def _setup_environ(debug=False):
         os.environ['PYTHONMALLOC'] = 'default'
         os.environ['PYTHONDEVMODE'] = ''
 
+    interesting_envs = {
+        k: os.environ[k]
+        for k in os.environ
+        if k.startswith(('PYTHON', 'GEVENT'))
+    }
+    widest_k = max(len(k) for k in interesting_envs)
+    for k, v in sorted(interesting_envs.items()):
+        util.log('%*s\t=\t%s', widest_k, k, v, color="debug")
+
 
 def main():
     # pylint:disable=too-many-locals,too-many-statements
     import argparse
     parser = argparse.ArgumentParser()
     parser.add_argument('--ignore')
-    parser.add_argument('--discover', action='store_true')
-    parser.add_argument('--full', action='store_true')
-    parser.add_argument('--config', default='known_failures.py')
-    parser.add_argument('--failfast', '-x', action='store_true')
-    parser.add_argument("--coverage", action="store_true")
-    parser.add_argument("--quiet", action="store_true", default=True)
+    parser.add_argument(
+        '--discover', action='store_true',
+        help="Only print the tests found."
+    )
+    parser.add_argument(
+        '--config', default='known_failures.py',
+        help="The path to the config file containing "
+        "FAILING_TESTS, IGNORED_TESTS and RUN_ALONE. "
+        "Defaults to %(default)s."
+    )
+    parser.add_argument(
+        "--coverage", action="store_true",
+        help="Enable coverage recording with coverage.py."
+    )
+    # TODO: Quiet and verbose should be mutually exclusive
+    parser.add_argument(
+        "--quiet", action="store_true", default=True,
+        help="Be quiet. Defaults to %(default)s. Also the "
+        "GEVENTTEST_QUIET environment variable."
+    )
     parser.add_argument("--verbose", action="store_false", dest='quiet')
-    parser.add_argument("--debug", action="store_true", default=False)
-    parser.add_argument("--package", default="gevent.tests")
+
+    parser.add_argument(
+        "--debug", action="store_true", default=False,
+        help="Enable debug settings. If the GEVENT_DEBUG environment variable is not set, "
+        "this sets it to 'debug'. This can also enable PYTHONTRACEMALLOC and the debug PYTHONMALLOC "
+        "allocators, if not already set. Defaults to %(default)s."
+     )
+
+    parser.add_argument(
+        "--package", default="gevent.tests",
+        help="Load tests from the given package. Defaults to %(default)s."
+    )
     parser.add_argument(
         "--processes", "-j", default=DEFAULT_NWORKERS, type=int,
         help="Use up to the given number of parallel processes to execute tests. "
@@ -714,9 +858,17 @@ def main():
                         'For example, "-u-network". GEVENTTEST_USE_RESOURCES is used '
                         'if no argument is given. To only use one resources, specify '
                         '"-unone,resource".')
-
     parser.add_argument("--travis-fold", metavar="MSG",
                         help="Emit Travis CI log fold markers around the output.")
+
+    fail_parser = parser.add_mutually_exclusive_group()
+    fail_parser.add_argument(
+        "--second-chance", action="store_true", default=False,
+        help="Give failed tests a second chance.")
+    fail_parser.add_argument(
+        '--failfast', '-x', action='store_true', default=False,
+        help="Stop running after the first failure.")
+
     parser.add_argument('tests', nargs='*')
     options = parser.parse_args()
     # options.use will be either None for not given, or a list
@@ -759,6 +911,8 @@ def main():
             # coverage files (which will have distinct suffixes because of parallel=true in .coveragerc
             # in this directory; makes them easier to combine and use with coverage report)
             os.environ['COVERAGE_FILE'] = os.path.abspath(".") + os.sep + ".coverage"
+            # XXX: Log this with color. Right now, it interferes (buffering) with other early
+            # output.
             print("Enabling coverage to", os.environ['COVERAGE_FILE'],
                   "with site", site_dir,
                   "and configuration file", cov_config)
@@ -770,7 +924,7 @@ def main():
     if options.config:
         config = {}
         options.config = _package_relative_filename(options.config, options.package)
-        with open(options.config) as f:
+        with open(options.config) as f: # pylint:disable=unspecified-encoding
             config_data = f.read()
         six.exec_(config_data, config)
         FAILING_TESTS = config['FAILING_TESTS']
@@ -806,6 +960,7 @@ def main():
             quiet=options.quiet,
             configured_run_alone_tests=RUN_ALONE,
             worker_count=options.processes,
+            second_chance=options.second_chance,
         )
 
         if options.travis_fold:
